# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

if __name__ == '__main__':
    import sys
    from constructor.main import main

    sys.exit(main())
